#import "MyOpenGLVideoView.h"

#define BailErr(x) {err = x; if(err != noErr) goto bail;}
#define BailIfNull(x) {gSeqGrab = x; if(gSeqGrab == nil) goto bail;}

static MyOpenGLVideoView	*myViewObject;	// our MyQuickDrawView object

@implementation MyOpenGLVideoView

//Initialize stuff at the beginning...
- (id) initWithCoder: (NSCoder *) coder
{
	self = [super initWithCoder: coder] ;
   
  //TextureSize
	glSize.width = 128;
	glSize.height = 128;
	backgroundSize.width = 512;
	backgroundSize.height = 512;
  //VideoSize
	sourceSize.width = 320;
    sourceSize.height = 240;
  //Status flags
	removeBackground = false;
	recordBackground = false;
	bluescreening = false;
	overlayYogi = true;
		
	[self initGLWindow ];
		
	
	saveOpenGLVideoViewObjectForCallback(self);
	
	//initialize and start all that QT-Stuff
	[ self startGrab: [self bounds] ];

	//create OpenCV Images
	source = cvCreateImage (sourceSize, IPL_DEPTH_8U, 4);
	
	workGL = cvCreateImage (glSize, IPL_DEPTH_8U, 4); 
	work = cvCreateImage (glSize, IPL_DEPTH_8U, 4);
	result = cvCreateImage (glSize, IPL_DEPTH_8U, 4);
	backgroundMean = cvCreateImage (glSize, IPL_DEPTH_32F, 4);
	backgroundVariance3 = cvCreateImage (glSize, IPL_DEPTH_32F, 4);
	backgroundSum = cvCreateImage (glSize, IPL_DEPTH_32S, 4);
	backgroundSquaredSum = cvCreateImage (glSize, IPL_DEPTH_32S, 4);
	runningSource = cvCreateImage(glSize, IPL_DEPTH_32F, 4);

	divisionSpeedup = sqrt(2*M_PI * 2*M_PI * 2*M_PI);

	return self ;
}

//set some properties for OpenGL
- (void) initGLWindow
{
	NSOpenGLPixelFormatAttribute attrs[] = 
	{
		NSOpenGLPFADoubleBuffer, //Our Wishlist for the OpenGL Context
		NSOpenGLPFAAccelerated,
		NSOpenGLPFADepthSize, 16,
		NSOpenGLPFAAlphaSize, 8,
		nil
	} ;
	
	NSOpenGLPixelFormat* pixFmt ;
	
	pixFmt = [[NSOpenGLPixelFormat alloc] initWithAttributes: attrs] ;
	[self setPixelFormat: pixFmt] ;
	
	//don't wait for vertical retrace to update image. (We have LCDs anyways...)
	long swapInterval = 0 ;
	[[self openGLContext]
        setValues: &swapInterval
	 forParameter: NSOpenGLCPSwapInterval ] ;	

	// allocate a texture name to store the video from iSight
	glGenTextures (1, &videoTexture);
	// select our current texture. All following Tex-Commands modify that texture
	glBindTexture( GL_TEXTURE_2D, videoTexture );
	
	// when texture area is small, bilinear filter the original (usually use MIPMaps!)
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	// when texture area is large, bilinear filter the original
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	
	// the texture wraps over at the edges (repeat)
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	
	// select way to mix texture with color, we'll just add the colors...
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	//build texture;
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, glSize.width, glSize.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);		

	
	
	
	// allocate a texture name to store the background
	glGenTextures (1, &backgroundTexture);
	// select our current texture. All following Tex-Commands modify that texture
	glBindTexture( GL_TEXTURE_2D, backgroundTexture );
	
	// when texture area is small, bilinear filter the original (usually use MIPMaps!)
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	// when texture area is large, bilinear filter the original
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	
	// the texture wraps over at the edges (repeat)
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	
	// select way to mix texture with color, we'll just add the colors...
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	//build texture;
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, backgroundSize.width, backgroundSize.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);		
}

//cleaning up all that mess.
- (void) dealloc
{
	[ self endGrab ];
	
	cvReleaseImage( &source );
	cvReleaseImage( &workGL );
	cvReleaseImage( &work );
	cvReleaseImage( &result );
	cvReleaseImage( &backgroundMean );
	cvReleaseImage( &backgroundVariance3 );
	cvReleaseImage( &backgroundSum );
	cvReleaseImage( &backgroundSquaredSum );
	cvReleaseImage( &runningSource );	
}


// Perform clean-up when we are finished recording
-(void)endGrab
{
    ComponentResult res;
    OSErr 			err;
	
    // kill our sequence grabber idle timer first
    [gMyTimer invalidate];
    [gMyTimer release];
    
    // stop recording
    SGStop(gSeqGrab);
	
    // end our decompression sequences
    err = CDSequenceEnd(gDecomSeq);
	
    // finally, close our sequence grabber component
    res = CloseComponent(gSeqGrab);
    
    // get rid of our gworld
    DisposeGWorld(offScreen);
}


//We got a new image from iSight, plz update display.
- (void)updateImage:(char*)data
{
	// do somwthing with videoTexture
	glBindTexture( GL_TEXTURE_2D, videoTexture );
	
	//glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, glSize.width, glSize.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);	
	//the following should do the same, just a little faster...
	glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, glSize.width, glSize.height, GL_RGBA, GL_UNSIGNED_BYTE, data);	
	
	//plz update display if you find the time for that... (e.g. it is allready drawing, so let it finish)
	//will call drawRect if it has time to do so...
	[self setNeedsDisplay: YES] ;
}



//This (overloaded) method does the actual drawing. 
//It is called everytime the texture changes or the mouse is dragged...
- (void) drawRect: (NSRect) bounds
{
	//Do some setup for camera
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//Setup Frustum
	glViewport (0,0, [self bounds].size.width, [self bounds].size.height );
	glOrtho( -1,1,
			 0,1,
			 -1,1);
	
	//Position the camera
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
    glClear( GL_COLOR_BUFFER_BIT ) ;
	
	glDisable ( GL_DEPTH_TEST );
	glDisable (GL_BLEND);

	//make a polygon with background painted onto it
	//if we do bluescreening, else just draw blue polygon
	if (bluescreening){
		glEnable ( GL_TEXTURE_2D );
		glColor4f ( 1.0, 1.0, 1.0, 1.0);
		glBindTexture( GL_TEXTURE_2D, backgroundTexture );
	}else{
		glDisable ( GL_TEXTURE_2D );
		glColor4f ( 0.0, 0.0, 1.0,1.0);
	}
	
	glBegin( GL_QUADS ) ;
	{       
		glTexCoord2f(1,1); glVertex3f( -1.0f, .0f, 0.0f ) ;
		glTexCoord2f(0,1); glVertex3f( 1.0f, 0.0f, 0.0f ) ;
		glTexCoord2f(0,0); glVertex3f( 1.0f, 1.0f, 0.0f ) ;
		glTexCoord2f(1,0); glVertex3f( -1.0f, 1.0f, 0.0f ) ;
	}
	glEnd() ;
	
	//make a polygon with video.
	//Enable Blending, else we won't see nothing but the video...
	glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glColor4f ( 1.0, 1.0, 1.0, 1.0);
	glEnable ( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, videoTexture );
	glBegin( GL_QUADS ) ;
	{       
		glTexCoord2f(1,1); glVertex3f( -1.0f, .0f, 0.0f ) ;
		glTexCoord2f(0,1); glVertex3f( 1.0f, 0.0f, 0.0f ) ;
		glTexCoord2f(0,0); glVertex3f( 1.0f, 1.0f, 0.0f ) ;
		glTexCoord2f(1,0); glVertex3f( -1.0f, 1.0f, 0.0f ) ;
	}
    glEnd() ;

	//done. As this is Double-Buffered, instruct system to draw at will. (long pipeline -> async!)
    [[self openGLContext] flushBuffer] ;
}







//Some Toggles to play with...
- (IBAction)toggleBackgroundRemoval:(id)sender
{
	if (removeBackground)
		removeBackground = false;
	else
		removeBackground = true;

	NSLog(@"Removal of Background Toggled");

}

- (IBAction)toggleRecordingOfBackground:(id)sender
{
	if (recordBackground){
		recordBackground = false;
		NSLog(@"Recording of Background Stopped with %d images added", nrOfBackgroundImages);
	}else{
		recordBackground = true;	
		NSLog(@"Recording of Background Started");
		nrOfBackgroundImages = 0;
	}
}

- (IBAction)toggleOverlay:(id)sender
{
	if (overlayYogi)
		overlayYogi = false;
	else
		overlayYogi = true;
	
	NSLog(@"Overlay Toggled");
}

- (IBAction)toggleBluesceening:(id)sender
{
	if (bluescreening)
		bluescreening = false;
	else{
		char * backgroundData = [[[ background image ] bestRepresentationForDevice : nil ] bitmapData ];
		
		if (!backgroundData){
			NSLog(@"Woops, got no data from background!");
			return;
		}
		
		CvSize tempSize;
		tempSize.width = [[[ background image ] bestRepresentationForDevice : nil ] size ].width;
		tempSize.height = [[[ background image ] bestRepresentationForDevice : nil ] size ].height;

		IplImage * temp = cvCreateImage(tempSize, IPL_DEPTH_8U, 4);
		NSLog(@"Build Temp-Image of %d x %d",tempSize.width,tempSize.height);
		
		int x,y, base4, base3;
		for (y = 0; y < tempSize.height; y++){ 
			for (x = 0; x < tempSize.width; x++){
				base3 = 3* (y * tempSize.width + x);
				base4 = 4* ((y+1) * tempSize.width - x-1);//mirror!
				temp->imageData[ base4 ] = backgroundData[ base3 ];
				temp->imageData[ base4+1 ] = backgroundData[ base3+1 ];
				temp->imageData[ base4+2 ] = backgroundData[ base3+2 ];
				temp->imageData[ base4+3 ] = 255;
			}
		}
		NSLog(@"Got Background Image.");


//GL OVERLAY
		IplImage * tempGL = cvCreateImage(backgroundSize, IPL_DEPTH_8U, 4);
		cvResize ( temp, tempGL, CV_INTER_LINEAR); //to have output in any case

		// do somwthing with videoTexture
		glBindTexture( GL_TEXTURE_2D, backgroundTexture );
		
		glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, backgroundSize.width, backgroundSize.height, GL_RGBA, GL_UNSIGNED_BYTE, tempGL->imageData);	

		NSLog(@"Uploaded into GL.");
		cvReleaseImage( &tempGL );
		cvReleaseImage( &temp );
		
		bluescreening = true;		
	}
	NSLog(@"Bluescreening Toggled");
}







//As QT is a fscking old mess, we have to tell the SeqenceGraber: "I have CPU-Cycles to spare
//"plz do something." Nobody needed Multitasking anyway...
-(void) sgIdleTimer:(id)sender
{
    OSErr err;

    err = SGIdle(gSeqGrab);
    if (err != noErr)
    {
		NSLog(@"fucked up within timer!");
		// ...to fix this we simply call SGStop and SGStartRecord again
        // hopefully fixing any ...		
        SGStop(gSeqGrab);
        SGStartRecord(gSeqGrab);
    }
}

//Create Offscreen Buffer and start all that QT-Mess
-(OSErr) startGrab:(NSRect)grabRect
{	
	OSErr	err = noErr;
    
    // initialize the movie toolbox
    err = EnterMovies();
    BailErr(err);
    
    // open the sequence grabber component and initialize it
    gSeqGrab = OpenDefaultComponent(SeqGrabComponentType, 0);
    BailIfNull(gSeqGrab);
    
    err = SGInitialize(gSeqGrab);
    BailErr(err);
    
	// specify the destination data reference for a record operation
	// tell it we're not making a movie
	// if the flag seqGrabDontMakeMovie is used, the sequence grabber still calls
	// your data function, but does not write any data to the movie file
	// writeType will always be set to seqGrabWriteAppend
    err = SGSetDataRef(gSeqGrab, 0, 0, seqGrabDontMakeMovie);
    BailErr(err);
	
	
    // create a new sequence grabber video channel
    err = SGNewChannel(gSeqGrab, VideoMediaType, &gSGChanVideo);
    BailErr(err);


	
	///////////////////////////
	//RESTORING PREFERENCES. CUT&PASTE FROM APPLE FAQ
	/////////////////////////
	UserData mySGVideoSettings = NULL;
		
		// get the settings using the key "sgVideoSettings"
		
		GetSettingsPreference(CFSTR("sgVideoSettings"), &mySGVideoSettings);
	
	if (mySGVideoSettings) {
		// use the saved settings preference to configure the SGChannel
		
		SGSetChannelSettings(gSeqGrab, gSGChanVideo,
							 mySGVideoSettings, 0);
		
		DisposeUserData(mySGVideoSettings);
		
	}
	
	
	// bring up the settings dialog to allow for further configuration
	
	err = SGSettingsDialog(gSeqGrab, gSGChanVideo, 0, nil, 0, nil, 0);
	BailErr(err);
	
	
	///////////////////////////
	//SAVING PREFERENCES. CUT&PASTE FROM APPLE FAQ
	/////////////////////////
	// get the SGChannel settings cofigured by the user
	SGGetChannelSettings(gSeqGrab, gSGChanVideo, &mySGVideoSettings, 0);
	// save the settings using the key "sgVideoSettings" 
	
	SaveSettingsPreference(CFSTR("sgVideoSettings"), mySGVideoSettings);
	
	DisposeUserData(mySGVideoSettings);
	////////////////////////
	
	
	//Tell preferred size (must be something nice! 640x480, 320x240, 176x144, all those pal, half-ntsc, ...
    gBoundsRect.top 	= 0;
    gBoundsRect.left 	= 0;
    gBoundsRect.bottom 	= sourceSize.height;
    gBoundsRect.right 	= sourceSize.width;
    err = SGSetChannelBounds(gSeqGrab, &gBoundsRect);

	NSLog (@"SeqGrab Channel Bounds: x = %d, y = %d",  gBoundsRect.right,  gBoundsRect.bottom); 

	
    // create the Offscreen GWorld
    err = QTNewGWorld(&offScreen,			// returned GWorld
    				  k32ARGBPixelFormat,		// pixel format (Note to self: k32RGBAPixelFormat olny works!)
    				  &gBoundsRect,				// bounding rectangle
    				  0,						// color table
    				  NULL,						// graphic device handle
    				  0);						// flags
    BailErr(err);
	
    // lock the pixmap and make sure it's locked because
    // we can't decompress into an unlocked PixMap
    if(!LockPixels(GetPortPixMap(offScreen)))
    {
    	BailErr(-1);
    }
	
	//marry Offscreen and SeqGrabber
    err = SGSetGWorld(gSeqGrab, offScreen, GetMainDevice());
    BailErr(err);
	
    // set the bounds for the channel
    err = SGSetChannelBounds(gSGChanVideo, &gBoundsRect);
    BailErr(err);
    
    // set the usage for our new video channel to avoid playthrough
    // note: we do not set seqGrabPlayDuringRecord because if you set this flag
    // the data from the channel may be played during the record operation,
    // if the destination buffer is onscreen. However, playing the
    // data may affect the quality of the recorded sequence by causing frames 
    // to be dropped...something we definitely want to avoid
    err = SGSetChannelUsage(gSGChanVideo, seqGrabRecord);
    BailErr(err);
    
    // The following function is called every time, an image is ready. Here we will
	// e.g. Sobelfilter the image.
    err = SGSetDataProc(gSeqGrab,NewSGDataUPP(&mySGDataProc),NULL);
    BailErr(err);
	
    /* lights...camera... */
    err = SGPrepare(gSeqGrab,false,true);
    BailErr(err);
    
    // start recording!!
    err = SGStartRecord(gSeqGrab);
    BailErr(err);
	
	// setup a timer to idle the sequence grabber... 
    gMyTimer = [[NSTimer scheduledTimerWithTimeInterval:0.01		// interval, 0.1 seconds might suffice.
												 target:self
											   selector:@selector(sgIdleTimer:)		// call this method
											   userInfo:nil
												repeats:YES] retain];					// repeat until we cancel it

	////////////
    // Camera frames are now put into some buffer. Possibly in compressed form. Ne now need to extract those
	// into the Offscreen Buffer
	///////////
	
	ImageDescriptionHandle	imageDesc = (ImageDescriptionHandle)NewHandle(0);
    
    // retrieve a channels current sample description, the channel returns a sample description that is
    // appropriate to the type of data being captured. Hopefully identical to _request_ for resolution earlier
    err = SGGetChannelSampleDescription(gSGChanVideo,(Handle)imageDesc);
    BailErr(err);
	
	NSLog (@"actual Sample-Size: x = %d, y = %d",  (*imageDesc)->width,  (*imageDesc)->height); 
		
    // begin the process of decompressing a sequence of frames
    // this is a set-up call and is only called once for the sequence - the ICM will interrogate different codecs
    // and construct a suitable decompression chain, as this is a time consuming process we don't want to do this
    // once per frame (eg. by using DecompressImage)
    // for more information see Ice Floe #8 http://developer.apple.com/quicktime/icefloe/dispatch008.html
    // the destination is specified as the GWorld
    err = DecompressSequenceBegin(&gDecomSeq,	// pointer to field to receive unique ID for sequence
								  imageDesc,	// handle to image description structure
								  offScreen,	// port for the DESTINATION image
								  NULL,			// graphics device handle, if port is set, set to NULL
								  NULL,			// source rectangle defining the portion of the image to decompress
								  NULL,			// transformation matrix
								  srcCopy,		// transfer mode specifier
								  NULL,			// clipping region in dest. coordinate system to use as a mask
								  0,					// flags
								  codecNormalQuality,   // accuracy in decompression
								  bestSpeedCodec);		// compressor identifier or special identifiers ie. bestSpeedCodec
    BailErr(err);
    
    DisposeHandle((Handle)imageDesc);
    imageDesc = NULL;
    
	BailErr(err);
	
bail:
		
		return err;
}

-(GWorldPtr)gworld
{
    return offScreen;
}

-(ImageSequence)decomSeq
{
    return gDecomSeq;
}

-(SGChannel)sgChanVideo
{
    return gSGChanVideo;
}

-(Rect)boundsRect
{
    return gBoundsRect;
}

-(IplImage *)getSource
{
	return source;
}


-(void)doImageProcessing
{
	unsigned int x;
//some pointers to get image data quickly
	unsigned char * wrk = (char*)(work->imageData);
	unsigned char * wrkGL = (unsigned char*)(workGL->imageData);
	signed int * sum = (signed int*)(backgroundSum->imageData);
	signed int * square = (signed int*)(backgroundSquaredSum->imageData);
	float * mean = (float*)(backgroundMean->imageData);
	float * variance = (float*)(backgroundVariance3->imageData);
	float * rSrc = (float*)(runningSource->imageData);
	unsigned char * res = (unsigned char*)(result->imageData);
	
	int s = [ gaussian intValue ];
	cvSmooth ( source, source, CV_GAUSSIAN, s, s );
	cvResize ( source, workGL, CV_INTER_LINEAR); //to have output in any case
	
	
	if (recordBackground){ //get Background for difference now!		
		
		unsigned char pixel;
		
		for (x = 0; x < backgroundSum->height * backgroundSum->width * 4;) {
			pixel = wrkGL [x];
			sum [x] += pixel;
			square [x] += pixel * pixel;
			x++;
			pixel = wrkGL [x];
			sum [x] += pixel;
			square [x] += pixel * pixel;
			x++;
			pixel = wrkGL [x];
			sum [x] += pixel;
			square [x] += pixel * pixel;
			x++;
			x++;
		}
		nrOfBackgroundImages++;
		
		if (nrOfBackgroundImages >= 128){
			recordBackground = false;
			removeBackground = true;
			
			NSLog(@"Recording of Background Automatically Stopped with %d images added", nrOfBackgroundImages);
			
			double diff;
			double min =  0.0001;
			
			for (x = 0; x < backgroundSum->height * backgroundSum->width * 4;) {
				double det = 0;
				mean[x] = sum[x] / nrOfBackgroundImages;
				diff = (square [x] / (nrOfBackgroundImages * 1.0)) - ( mean [ x ] * mean [ x ]);
if (abs(diff) == 0.0)
   diff = min;
				variance [ x ] = -0.5 / sqrt( abs(diff) );
				det = sqrt( abs(diff) );
				sum[x] = 0;
				square[x] = 0;
				x++;
				mean[x] = sum[x] / nrOfBackgroundImages;
				diff = (square [x] / (nrOfBackgroundImages * 1.0)) - ( mean [ x ] * mean [ x ]);
if (abs(diff) == 0.0)
   diff = min;
				variance [ x ] = -0.5 / sqrt( abs(diff) );
				det *= sqrt( abs(diff) );
				sum[x] = 0;
				square[x] = 0;
				x++;
				mean[x] = sum[x] / nrOfBackgroundImages;
				diff = (square [x] / (nrOfBackgroundImages * 1.0)) - ( mean [ x ] * mean [ x ]);
if (abs(diff) == 0.0)
   diff = min;
				variance [ x ] = -0.5 / sqrt( abs(diff) );
				det *= sqrt( abs(diff) );
				sum[x] = 0;
				square[x] = 0;
				x++;
				variance [x] =  divisionSpeedup * sqrt(det) ; //log Multiplicator in Normaldistribution -> add to compute ln(p)!
				if (variance [x] == 0.0)  //beware of "zero" variance -> -inf -> permanent blue!
					variance [x] = min;
				variance [x] = log( variance[ x ] );
				x++;
			}
			
			NSLog(@"Background Mean and Variance computed!");
		}
	}
	
	if (removeBackground){ //compute Difference Image	
		float diff0,diff1,diff2;
		
		int dnoise = [ deNoise intValue ];
		a = [ alpha floatValue ]; //read out spatial smoothing
		float b = [ threshold floatValue ];
				
		if (a < 1.0){
			for (x = 0; x < backgroundSum->height * backgroundSum->width*4; x+=4) {
				rSrc [x] = a * wrkGL[x] + (1-a)*rSrc[x];
				rSrc [x+1] = a * wrkGL[x+1] + (1-a)*rSrc[x+1];
				rSrc [x+2] = a * wrkGL[x+2] + (1-a)*rSrc[x+2];
				//real probability modeling
				diff0 = mean[x] - rSrc [x];
				diff1 = mean[x+1] - rSrc [x+1];
				diff2 = mean[x+2] - rSrc [x+2];
				
				float e = diff0 * variance[x] * diff0; //exponent of e functional
				e += diff1 * variance[x+1] * diff1;
				e += diff2 * variance[x+2] * diff2;
				float lnProb = e - variance[x+3];
				
				res [ x ] = 0;
				res [ x+1 ] = 0;
				res [ x+2 ] = 0;
				res [ x+3 ] = 0;
				
				if (lnProb < b){ //lnProb = Probaility for pixel beeing background
					res [ x ] = 255;
					res [ x+1 ] = 255;
					res [ x+2 ] = 255;
					res [ x+3 ] = 255;
				}
			}
		}else{
			for (x = 0; x < backgroundSum->height * backgroundSum->width*4; x+=4) {
				//real probability modeling
				diff0 = mean[x] - wrkGL [x];
				diff1 = mean[x+1] - wrkGL [x+1];
				diff2 = mean[x+2] - wrkGL [x+2];
				
				float e = diff0 * variance[x] * diff0; //exponent of e functional
				e += diff1 * variance[x+1] * diff1;
				e += diff2 * variance[x+2] * diff2;
				float lnProb = e - variance[x+3];
				
				res [ x ] = 0;
				res [ x+1 ] = 0;
				res [ x+2 ] = 0;
				res [ x+3 ] = 0;
				
				if (lnProb < b){
					res [ x ] = 255;
					res [ x+1 ] = 255;
					res [ x+2 ] = 255;
					res [ x+3 ] = 255;
				}
			}
		}
		
		if (dnoise == 1){
			cvDilate ( result, result, NULL, 2);
			cvErode ( result, result, NULL, 2);
		}
		if (dnoise == 2){
			cvDilate ( result, result, NULL, 2);
			cvErode ( result, result, NULL, 3);
			cvDilate ( result, result, NULL, 1);
		}
		
		if (overlayYogi){
			for (x = 0; x < backgroundSum->height * backgroundSum->width*4; x+=4) {
				if (res [ x ] | res [ x+1 ] | res [ x+2 ]){ //copy video into sementation-map
					res [ x ] = wrkGL [x];
					res [ x+1 ] = wrkGL [x+1];
					res [ x+2 ] = wrkGL [x+2];
				}else{
					float c = 0.08; //background addaption to changes;
					mean [x] = c * wrkGL[x] + (1-c)*mean[x];
					mean [x+1] = c * wrkGL[x+1] + (1-c)*mean[x+1];
					mean [x+2] = c * wrkGL[x+2] + (1-c)*mean[x+2];
				}
			}
		}
		
		[ self updateImage: result->imageData ];
		return;
	}
	
	[ self updateImage: workGL->imageData ];
	
}

@end



/* ---------------------------------------------------------------------- */
/* sequence grabber data procedure - this is where the work is done       */
/* ---------------------------------------------------------------------- */
/* mySGDataProc - the sequence grabber calls the data function whenever
any of the grabbers channels write digitized data to the destination movie file.

NOTE: We really mean any, if you have an audio and video channel then the DataProc will
be called for either channel whenever data has been captured. Be sure to check which
channel is being passed in. In this example we never create an audio channel so we know
we're always dealing with video.

This data function decompresses captured video data into an offscreen GWorld,
then transfers the frame to an onscreen window.

For more information refer to Inside Macintosh: QuickTime Components, page 5-120
c - the channel component that is writing the digitized data.
p - a pointer to the digitized data.
len - the number of bytes of digitized data.
offset - a pointer to a field that may specify where you are to write the digitized data,
and that is to receive a value indicating where you wrote the data.
chRefCon - per channel reference constant specified using SGSetChannelRefCon.
time	- the starting time of the data, in the channels time scale.
writeType - the type of write operation being performed.
seqGrabWriteAppend - Append new data.
seqGrabWriteReserve - Do not write data. Instead, reserve space for the amount of data
specified in the len parameter.
seqGrabWriteFill - Write data into the location specified by offset. Used to fill the space
previously reserved with seqGrabWriteReserve. The Sequence Grabber may
call the DataProc several times to fill a single reserved location.
refCon - the reference constant you specified when you assigned your data function to the sequence grabber.
*/


pascal OSErr mySGDataProc(SGChannel c, 
						  Ptr p,
						  long len,
						  long *offset,
						  long chRefCon,
						  TimeValue time,
						  short writeType, 
						  long refCon)
{
#pragma unused(offset,chRefCon,time,writeType)
    CodecFlags 		ignore;
	ComponentResult err = noErr;

    if([myViewObject gworld]) 
    {			
        // decompress a frame into the GWorld - can queue a frame for async decompression when passed in a completion proc
        // once the image is in the GWorld it can be manipulated at will
        err = DecompressSequenceFrameS([myViewObject decomSeq],	// sequence ID returned by DecompressSequenceBegin
									   p,						// pointer to compressed image data
									   len,					// size of the buffer
									   0,						// in flags
									   &ignore,				// out flags
									   NULL);					// async completion proc
        BailErr(err);
				
        // ******  IMAGE IS NOW IN THE GWORLD ****** //
	}
	
	// Transform Pixmap for openCV...
	PixMapHandle	pix = GetGWorldPixMap([myViewObject gworld]);
	Rect		bounds;
	
	GetPortBounds([myViewObject gworld], &bounds);

	unsigned int y, srcBase, trgBase;
	unsigned int srcRowSize = (unsigned int)GetPixRowBytes(pix);
	unsigned int srcPixBase = (unsigned int)GetPixBaseAddr(pix);
	
	unsigned int trgRowSize = [myViewObject getSource ] -> widthStep;
	unsigned int trgPixBase = (unsigned int)([myViewObject getSource ] -> imageData);
	
    for (y = 0; y < bounds.bottom; y++)
    {
        srcBase = srcPixBase + y * srcRowSize + 1; //+1 due to ARGB vs RGBA!
        trgBase = trgPixBase + y * trgRowSize;
		memcpy ((char*)trgBase, (char*)srcBase, trgRowSize);
    }

	
	[ myViewObject doImageProcessing ];

bail:
		
		return err;	


}

//////////
//
// saveQDViewObjectForCallback
//
// This routine stores a reference to our MyQuickDrawView object. We'll
// need this so we can call into methods in this class from outside the
// implementation of the class methods (specifically, from our SGDataProc
// C routine above).
//
//////////

void saveOpenGLVideoViewObjectForCallback(void *theObject)
{
	myViewObject = (MyOpenGLVideoView *)theObject;
}








////////////////////////////////
//Functions to store Video-Preferences. I do NOT know exactly what those do...
////////////////////////////////



// GetSettingsPreference

//    Returns a preference for a specified key as QuickTime UserData

// It is your responsibility to dispose of the returned UserData

OSErr GetSettingsPreference(CFStringRef inKey, UserData *outUserData)

{
	
	CFPropertyListRef theCFSettings;
	
	Handle            theHandle = NULL;
	
	UserData          theUserData = NULL;
	
	OSErr             err = paramErr;
	
	
	
	// read the new setttings from our preferences
	
	theCFSettings = CFPreferencesCopyAppValue(inKey,
											  kCFPreferencesCurrentApplication);
	
	if (theCFSettings) {
		err = PtrToHand(CFDataGetBytePtr((CFDataRef)theCFSettings), &theHandle,
						CFDataGetLength((CFDataRef)theCFSettings));
		
        
		
		CFRelease(theCFSettings);
		
		if (theHandle) {
			
			err = NewUserDataFromHandle(theHandle, &theUserData);
			
			if (theUserData) {
				*outUserData = theUserData;
			}
			
			DisposeHandle(theHandle);
			
		}
		
	}
	
	
	return err;
	
}



// SaveSettingsPreference

//    Saves a preference for a specified key from QuickTime UserData

OSErr SaveSettingsPreference(CFStringRef inKey, UserData inUserData)

{
	
	CFDataRef theCFSettings;
	
	Handle    hSettings;
	
	OSErr     err;
	
    
	
	if (NULL == inUserData) return paramErr;
	
    
	
	hSettings = NewHandle(0);
	
	err = MemError();
	
    
	
	if (noErr == err) {
		
		err = PutUserDataIntoHandle(inUserData, hSettings); 
		
        
		
		if (noErr == err) {
			
			HLock(hSettings);
			
			
			
			theCFSettings = CFDataCreate(kCFAllocatorDefault,
										 (UInt8 *)*hSettings,
										 GetHandleSize(hSettings));
			if (theCFSettings) {
				
				CFPreferencesSetAppValue(inKey, theCFSettings,
										 kCFPreferencesCurrentApplication);
				
				CFPreferencesAppSynchronize(kCFPreferencesCurrentApplication);
				
				CFRelease(theCFSettings);
				
			}
			
		}
		
		
		DisposeHandle(hSettings);
		
	}
	
	
	return err;
	
}

